<?php
# @Author: Aaron Bowie
# @Date:   Saturday, September 16th 2017, 12:09:42 am
# @Email:  support@weareag.co.uk
# @Filename: ag-woocommerce-sagepay-class.php
# @Last modified by:   Aaron Bowie
# @Last modified time: Saturday, November 11th 2017, 8:27:31 pm


defined('ABSPATH') or die("No script kiddies please!");


/*-----------------------------------------------------------------------------------*/
/*  fraud status - TODO
/*-----------------------------------------------------------------------------------*/

//add_action( 'init', 'register_sagepay_fraud_order_status' );

function register_sagepay_fraud_order_status() {
    register_post_status(
        'wc-fraud-detection', array(
            'label'                     => _x( 'Fraud detection', 'Order status', 'ag-sagepay-direct' ),
            'public'                    => true,
            'exclude_from_search'       => false,
            'show_in_admin_all_list'    => true,
            'show_in_admin_status_list' => true,
            'label_count'               => _n_noop( 'Fraud Screening <span class="count">(%s)</span>', 'Fraud Screening <span class="count">(%s)</span>', 'ag-sagepay-direct' )
        ) );
}

//add_filter( 'wc_order_statuses', 'new_sagepay_order_statuses' );

function new_sagepay_order_statuses( $order_statuses ) {
    //$order_statuses['wc-authorised']    = _x( 'Authorised', 'Order status', 'ag-sagepay-direct' );
    $order_statuses['wc-fraud-detection']  = _x( 'Fraud detection', 'Order status', 'ag-sagepay-direct' );

    return $order_statuses;
}

/*-----------------------------------------------------------------------------------*/
/*  Set class
/*-----------------------------------------------------------------------------------*/

class ag_sagepay_direct_checkout extends WC_Payment_Gateway {

// Setup our Gateway's id, description and other values
function __construct() {

	$this->id = "ag_sagepay_direct";
	$this->method_title = __( "AG SagePay Direct", 'ag-sagepay-direct' );
  // The description for this Payment Gateway, shown on the actual Payment options page on the backend
  $license = get_option('AGD_SagePay_Key');
  $websiteurl = get_option ('siteurl');
  $api_params = array(
    'slm_action' => 'slm_check',
    'secret_key' => AGD_KEY,
    'license_key' => $license,
  );
  $query = esc_url_raw(add_query_arg($api_params, AGD));
  $response = wp_remote_get($query, array('timeout' => 20, 'sslverify' => false));
  $license_data = json_decode(wp_remote_retrieve_body($response));
  if($license_data->result == 'success'){
    $this->method_description =  __( "SagePay Direct payment gateway plugin for WooCommerce", 'ag-sagepay-direct' );
  } else {
    $this->method_description = __( 'Please enter the license key for this product to activate it. You were given a license key when you purchased this item. <br /><strong>You can do this on the <a href="'. $websiteurl.'/wp-admin/admin.php?page=AG-Woo-SagePay%2Fag-woocommerce-sagepay.php">AG License</a> menu item on the left</strong>.', 'ag-sagepay-direct' );
  }

	$this->title = __( "SagePay Direct", 'ag-sagepay-direct' );
	$this->icon = plugins_url('AG-Woo-SagePay/img/sagepay-logo.png');
	$this->has_fields = true;
	// Supports
	$this->supports = array( 'default_credit_card_form', 'refunds' );
  // actions
  //add_action('init', array( $this, 'successful_request' ));
  add_action('woocommerce_api_woocommerce_ag_sagepay_direct', array( $this, 'successful_request' ));
  add_action('woocommerce_receipt_ag_sagepay_direct', array( $this, 'receipt_page' ));

  if($license_data->result == 'success'){
    $this->init_form_fields();
    $this->init_settings();
  }

	foreach ( $this->settings as $setting_key => $value ) {
		$this->$setting_key = $value;
	}

	add_action( 'admin_notices', array( $this,	'do_ssl_check' ) );

	// Save settings
	if ( is_admin() ) {
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
	}
} // End __construct()

public function payment_fields() {
  if ( $description = $this->get_description() ) {
    echo wpautop( wptexturize( $description ) );
  }

  if ( $this->supports( 'default_credit_card_form' ) ) {
    $this->credit_card_form();
  }
}

/*-----------------------------------------------------------------------------------*/
/*  Setting form
/*-----------------------------------------------------------------------------------*/
public function credit_card_form( $args = array(), $fields = array() ) {

  wp_enqueue_script( 'wc-credit-card-form' );

  if( $this->environment == "yes" ) { ?>
    <p class="form-row form-row-wide"><br /><strong>TEST MODE ACTIVE</strong><br />In test mode you can run test payments using card type VISA, card number <strong>4929 0000 0000 6</strong> with any CVC number and a valid expiration date. Check SagePay doc's for more cards to try <a target="_blank" href="https://www.sagepay.co.uk/support/test-your-integration/process-a-test-transaction">here</a>.<br /><br />If 3DSecure is enabled on the account you can test this also, the password for testing is password.<br /><br /></p>
  <?php }?>
  <p class="form-row form-row-wide">
    <label for="ag_sagepay_direct-card-type" style="font-size: .875em; ">Card type <span class="required">*</span></label>
    <select name="ag_sagepay_direct-card-type" class="ag_sagepay_direct-card-type">

      <?php
      $card_options = '<option value = "0">Select Card Type</option>';
      foreach ( $this->cardtypes as  $value ) {
        $card_options .= '<option value="' . $value . '">' . $value . '</option>';
      }
      echo $card_options; ?>

      </select>
  </p>
  <?php

  $default_args = array(
    'fields_have_names' => true,
  );

  $args = wp_parse_args( $args, apply_filters( 'woocommerce_credit_card_form_args', $default_args, $this->id ) );

  $default_fields = array(
    'card-number-field' => '<p class="form-row form-row-wide">
      <label for="' . esc_attr( $this->id ) . '-card-number">' . __( 'Card Number', 'woocommerce' ) . ' <span class="required">*</span></label>
      <input id="' . esc_attr( $this->id ) . '-card-number" class="input-text wc-credit-card-form-card-number" type="text" maxlength="20" autocomplete="off" placeholder="•••• •••• •••• ••••" name="' . ( $args['fields_have_names'] ? $this->id . '-card-number' : '' ) . '" />
    </p>',
    'card-expiry-field' => '<p class="form-row form-row-first">
      <label for="' . esc_attr( $this->id ) . '-card-expiry">' . __( 'Expiry (MM/YY)', 'woocommerce' ) . ' <span class="required">*</span></label>
      <input id="' . esc_attr( $this->id ) . '-card-expiry" class="input-text wc-credit-card-form-card-expiry" type="text" autocomplete="off" placeholder="' . __( 'MM / YY', 'woocommerce' ) . '" name="' . ( $args['fields_have_names'] ? $this->id . '-card-expiry' : '' ) . '" />
    </p>',
    'card-cvc-field' => '<p class="form-row form-row-last">
      <label for="' . esc_attr( $this->id ) . '-card-cvc">' . __( 'CVC', 'woocommerce' ) . ' <span class="required">*</span></label>
      <input id="' . esc_attr( $this->id ) . '-card-cvc" class="input-text wc-credit-card-form-card-cvc" type="text" autocomplete="off" placeholder="' . __( 'CVC', 'woocommerce' ) . '" name="' . ( $args['fields_have_names'] ? $this->id . '-card-cvc' : '' ) . '" />
    </p>'
  );

  $fields = wp_parse_args( $fields, apply_filters( 'woocommerce_credit_card_form_fields', $default_fields, $this->id ) );
  ?>
  <fieldset id="<?php echo $this->id; ?>-cc-form">
    <?php do_action( 'woocommerce_credit_card_form_start', $this->id ); ?>
    <?php
      foreach ( $fields as $field ) {
        echo $field;
      }
    ?>
    <?php do_action( 'woocommerce_credit_card_form_end', $this->id ); ?>
    <div class="clear"></div>
  </fieldset>
  <?php
}



/*-----------------------------------------------------------------------------------*/
/*  Build setting fields
/*-----------------------------------------------------------------------------------*/

public function init_form_fields() {

  $this->sage_cardtypes = apply_filters( 'ag_sagepaydirect_cardtypes', array(
    'MasterCard'		    => __( 'MasterCard Credit', 'ag-sagepay-direct' ),
    'MasterCard Debit'	=> __( 'MasterCard Debit', 'ag-sagepay-direct' ),
    'Visa'				      => __( 'Visa Credit', 'ag-sagepay-direct' ),
    'Visa Debit'		    => __( 'Visa Debit', 'ag-sagepay-direct' ),
    'Discover'			    => __( 'Discover', 'ag-sagepay-direct' ),
    'American Express' 	=> __( 'American Express', 'ag-sagepay-direct' ),
    'Maestro'			      => __( 'Maestro', 'ag-sagepay-direct' ),
    'JCB'				        => __( 'JCB', 'ag-sagepay-direct' ),
    'Laser'				      => __( 'Laser', 'ag-sagepay-direct' ),
    //'PayPal'				  => __( 'PayPal', 'ag-sagepay-direct' ),
  ) );

  $this->form_fields = array(
		'enabled'   => array(
			'title'		=> __( 'Enable / Disable', 'ag-sagepay-direct' ),
			'label'		=> __( 'Enable this payment gateway', 'ag-sagepay-direct' ),
			'type'		=> 'checkbox',
			'default'	=> 'no',
		),
		'title' => array(
			'title'		=> __( 'Title', 'ag-sagepay-direct' ),
			'type'		=> 'text',
			'desc_tip'	=> __( 'Payment title the customer will see during the checkout process.', 'ag-sagepay-direct' ),
			'default'	=> __( 'Credit card', 'ag-sagepay-direct' ),
		),
		'description' => array(
			'title'		  => __( 'Description', 'ag-sagepay-direct' ),
			'type'		  => 'textarea',
			'desc_tip'	=> __( 'Payment description the customer will see during the checkout process.', 'ag-sagepay-direct' ),
			'default'	  => __( 'Pay securely using your credit card.', 'ag-sagepay-direct' ),
			'css'		    => 'max-width:350px;'
		),
    'vendor' => array(
      'title'		  => __( 'SagePay Direct Vendor Name', 'ag-sagepay-direct' ),
      'type'		  => 'text',
      'desc_tip'	=> __( 'This is the Vendor Name provided by SagePay Direct when you signed up for an account.', 'ag-sagepay-direct' ),
    ),

    'cardtypes'	=> array(
      'title' 		=> __( 'Accepted Cards', 'woocommerce_sagepaydirect' ),
      'type' 			=> 'multiselect',
      'class'			=> 'chosen_select',
      'css'         	=> 'width: 350px;',
      'description' 	=> __( 'Select which card types to accept.', 'woocommerce_sagepaydirect' ),
      'default' 		=> '',
      'options' 		=> $this->sage_cardtypes,
    ),

    'transtype' => array(
      'title' => __('Transition Type', 'ag-sagepay-direct'),
      'type' => 'select',
      'css'  => 'height: 35px;',
      'options' => array(
        'PAYMENT' => __('Payment', 'ag-sagepay-direct'),
        //'DEFFERRED' => __('Deferred', 'ag-sagepay-direct'),
        //'AUTHENTICATE' => __('Authenticate', 'ag-sagepay-direct')
      ),
      'description' => __('Select Payment, Deferred or Authenticated.', 'ag-sagepay-direct')
    ),


		'environment' => array(
			'title'		=> __( 'SagePay Direct Test Mode', 'ag-sagepay-direct' ),
			'label'		=> __( 'Enable Test Mode', 'ag-sagepay-direct' ),
			'type'		=> 'checkbox',
			'description' => __( 'Place the payment gateway in test mode.', 'ag-sagepay-direct' ),
			'default'	=> 'no',
		),

	);
}

/*-----------------------------------------------------------------------------------*/
/*  admin notices for SSL and test mode.
/*-----------------------------------------------------------------------------------*/
public function do_ssl_check() {
	if( is_ssl() == false ) {
			echo "<div class=\"error\"><p>". sprintf( __( "<strong>%s</strong> is enabled and WooCommerce is not forcing the SSL certificate on your checkout page. Please ensure that you have a valid SSL certificate and that you are <a href=\"%s\">forcing the checkout pages to be secured.</a>" ), $this->method_title, admin_url( 'admin.php?page=wc-settings&tab=checkout' ) ) ."</p></div>";
	}
  if( $this->environment == "yes" ) {
    echo "<div class=\"update-nag\"><p>". sprintf( __( "<strong>%s</strong> is enabled and in test mode. Dont forget to switch off test mode when you have ran some test payments." ), $this->method_title, admin_url( 'admin.php?page=wc-settings&tab=checkout' ) ) ."</p></div>";
  }
}



public function generate_sagepaydirect_form($order_id) {
    global $woocommerce;

    $order = new WC_Order($order_id);

    $sagepaydirect_args = array_merge(
        array(
            'PaReq'    => WC()->session->get('sagepay_pareq'),
            'MD'        => WC()->session->get('sagepay_md'),
            'TermUrl'    => add_query_arg('wc-api', 'woocommerce_ag_sagepay_direct', home_url('/'))
        )
    );

    $sagepaydirect_args_array = array();

    foreach ($sagepaydirect_args as $key => $value) {
        $sagepaydirect_args_array[] = '<input type="hidden" name="'.esc_attr($key).'" value="'.esc_attr($value).'" />';
    }

    wc_enqueue_js('
    jQuery("body").block({
      message: "<img src=\"'.esc_url($woocommerce->plugin_url()).'/assets/images/ajax-loader.gif\" alt=\"Redirecting...\" style=\"float:left; margin-right: 10px;\" />'.__('Thank you for your order. We are now redirecting you to verify your card.', 'ag-sagepay-direct').'",
      overlayCSS:
      {
        background: "#fff",
        opacity: 0.6
      },
      css:
      {
        padding:        20,
        textAlign:      "center",
        color:          "#555",
        border:         "3px solid #aaa",
        backgroundColor:"#fff",
        cursor:         "wait",
        lineHeight:		"32px"
      }
    });
    jQuery("#submit_sagepaydirect_payment_form").click();
');

    return '<form action="'.esc_url(WC()->session->get('sagepay_acsurl')).'" method="post" id="sagepaydirect_payment_form">' . implode('', $sagepaydirect_args_array) . '<input type="submit" class="button-alt" id="submit_sagepaydirect_payment_form" value="'.__('Submit', 'ag-sagepay-direct').'" /> <a class="button cancel" href="'.esc_url($order->get_cancel_order_url()).'">'.__('Cancel order &amp; restore cart', 'ag-sagepay-direct').'</a>
</form>';
}




public function process_payment($order_id) {
    global $woocommerce;

    $order = new WC_Order($order_id);

    $credit_card = str_replace( array(' ', '-' ), '', $_POST['ag_sagepay_direct-card-number'] );
    $ccexp_expiry = str_replace( array( '/', ' '), '', $_POST['ag_sagepay_direct-card-expiry'] );

    $order_item = $order->get_items();
  	foreach( $order_item as $product ) {
  		$prodct_name[] = $product['name']." x".$product['qty'];
  	}
  	$basket = implode( ',', $prodct_name );

    $basket .= ':Order Total:---:---:---:---:'.$order->order_total;
    $basket = $item_loop.':'.$basket;

    $time_stamp = date("ymdHis");
    $orderid = $this->vendor . "-" . $time_stamp . "-" . $order_id;

    $ag_spd_arg['ReferrerID']            = '495E035E-B1B8-4BF7-82F0-5DEE0D6A380C';
    $ag_spd_arg['Amount']                = $order->order_total;
    $ag_spd_arg['CustomerEMail']         = $order->billing_email;
    $ag_spd_arg['BillingFirstnames']     = $order->billing_first_name;
    $ag_spd_arg['BillingSurname']        = $order->billing_last_name;
    $ag_spd_arg['BillingAddress1']       = $order->billing_address_1;
    $ag_spd_arg['BillingAddress2']       = $order->billing_address_2;
    $ag_spd_arg['BillingCity']           = $order->billing_city;

    if ($order->billing_country == 'US') {
        $ag_spd_arg['BillingState']      = $order->billing_state;
    } else {
        $ag_spd_arg['BillingState']      = '';
    }

    $ag_spd_arg['BillingPostCode']       = $order->billing_postcode;
    $ag_spd_arg['BillingCountry']        = $order->billing_country;
    $ag_spd_arg['BillingPhone']          = $order->billing_phone;

    //if ($this->cart_has_virtual_product() == true || $this->send_shipping == 'yes') {
    //    $ag_spd_arg['DeliverySurname']        = $order->get_billing_last_name();
    //    $ag_spd_arg['DeliveryFirstnames']    = $order->get_billing_first_name();
    //    $ag_spd_arg['DeliveryAddress1']    = $order->get_billing_address_1();
    //    $ag_spd_arg['DeliveryAddress2']    = $order->get_billing_address_2();
    //    $ag_spd_arg['DeliveryCity']        = $order->get_billing_city();

    //    if ($order->get_billing_country() == 'US') {
    //        $ag_spd_arg['DeliveryState']    = $order->get_billing_state();
    //    } else {
    //        $ag_spd_arg['DeliveryState']    = '';
    //    }

    //    $ag_spd_arg['DeliveryPostCode']    = $order->get_billing_postcode();
    //    $ag_spd_arg['DeliveryCountry']        = $order->get_billing_country();
    //} else {
        $ag_spd_arg['DeliveryFirstnames']  = $order->shipping_first_name;
        $ag_spd_arg['DeliverySurname']     = $order->shipping_last_name;
        $ag_spd_arg['DeliveryAddress1']    = $order->shipping_address_1;
        $ag_spd_arg['DeliveryAddress2']    = $order->shipping_address_2;
        $ag_spd_arg['DeliveryCity']        = $order->shipping_city;
        if ($order->shipping_country == 'US') {
            $ag_spd_arg['DeliveryState']   = $order->shipping_state;
        } else {
            $ag_spd_arg['DeliveryState']   = '';
        }
        $ag_spd_arg['DeliveryPostCode']    = $order->shipping_postcode;
        $ag_spd_arg['DeliveryCountry']     = $order->shipping_country;
    //}

    $ag_spd_arg['DeliveryPhone']           = $order->shipping_phone;
    $ag_spd_arg['CardHolder']              = $customer_order->billing_first_name . ' ' . $customer_order->billing_last_name;
    $ag_spd_arg['CardNumber']              = $credit_card;
    $ag_spd_arg['ExpiryDate']              = $ccexp_expiry;
    $ag_spd_arg['CV2']                     = $_POST['ag_sagepay_direct-card-cvc'];
    $ag_spd_arg['CardType']                = $_POST['ag_sagepay_direct-card-type'];
    $ag_spd_arg['VPSProtocol']             = "3.00";
    $ag_spd_arg['Vendor']                  = $this->vendor;
    $ag_spd_arg['Description']             = sprintf(__('Order #%s', 'ag-sagepay'), ltrim($order->get_order_number(), '#'));
    $ag_spd_arg['Currency']                = get_woocommerce_currency();
    $ag_spd_arg['TxType']                  = $this->transtype;
    $ag_spd_arg['VendorTxCode']            = $orderid;

    $post_values = "";
    foreach ($ag_spd_arg as $key => $value) {
        $post_values .= "$key=" . urlencode($value) . "&";
    }
    $post_values = rtrim($post_values, "& ");

    // Are we testing right now or is it a real transaction
    $environment = ( $this->environment == "yes" ) ? 'TRUE' : 'FALSE';

    // Decide which URL to post to
    if( is_ssl()) {
    	$environment_url = ( "FALSE" == $environment )
    						 ? 'https://live.sagepay.com/gateway/service/vspdirect-register.vsp'
    						 : 'https://test.sagepay.com/gateway/service/vspdirect-register.vsp';
    } else {
      $environment_url = 'https://test.sagepay.com/gateway/service/vspdirect-register.vsp';
    }

    $result = wp_remote_post($environment_url, array(
      'body' => $post_values,
      'method' => 'POST',
      'sslverify' => false
    ));



    WC()->session->set('sagepay_vtc', $orderid);
    WC()->session->set('sagepay_oid', $order_id);

    if (!is_wp_error($result) && $result['response']['code'] >= 200 && $result['response']['code'] < 300) {
        $sagepay_response = array();
        $lines = preg_split('/\r\n|\r|\n/', $result['body']);
        foreach ($lines as $line) {
            $key_value = preg_split('/=/', $line, 2);
            if (count($key_value) > 1) {
                $sagepay_response[trim($key_value[0])] = trim($key_value[1]);
            }
        }

        if (isset($sagepay_response['Status'])) {
            update_post_meta($order->id, 'Status', $sagepay_response['Status']);
        }
        if (isset($sagepay_response['StatusDetail'])) {
            update_post_meta($order->id, 'StatusDetail', $sagepay_response['StatusDetail']);
        }
        if (isset($sagepay_response['VPSTxId'])) {
            update_post_meta($order->id, 'VPSTxId', $sagepay_response['VPSTxId']);
        }
        if (isset($sagepay_response['CAVV'])) {
            update_post_meta($order->id, 'CAVV', $sagepay_response['CAVV']);
        }
        if (isset($sagepay_response['SecurityKey'])) {
            update_post_meta($order->id, 'SecurityKey', $sagepay_response['SecurityKey']);
        }
        if (isset($sagepay_response['TxAuthNo'])) {
            update_post_meta($order->id, 'TxAuthNo', $sagepay_response['TxAuthNo']);
        }
        if (isset($sagepay_response['AVSCV2'])) {
            update_post_meta($order->id, 'AVSCV2', $sagepay_response['AVSCV2']);
        }
        if (isset($sagepay_response['AddressResult'])) {
            update_post_meta($order->id, 'AddressResult', $sagepay_response['AddressResult']);
        }
        if (isset($sagepay_response['PostCodeResult'])) {
            update_post_meta($order->id, 'PostCodeResult', $sagepay_response['PostCodeResult']);
        }
        if (isset($sagepay_response['CV2Result'])) {
            update_post_meta($order->id, 'CV2Result', $sagepay_response['CV2Result']);
        }
        if (isset($sagepay_response['3DSecureStatus'])) {
            update_post_meta($order->id, '3DSecureStatus', $sagepay_response['3DSecureStatus']);
        }
        /*
        if (WC()->session->get('sagepay_vtc') != '') {
            update_post_meta($order->id, 'VendorTxCode', WC()->session->get('sagepay_vtc'));
        }
        */
        if ($sagepay_response['Status'] == "OK" || $sagepay_response['Status'] == "REGISTERED" || $sagepay_response['Status'] == "AUTHENTICATED") {
            $order->add_order_note(__('SagePay Direct Payment Completed.', 'ag-sagepay-direct'));
            $order->payment_complete();
            $redirect_url = $this->get_return_url($order);

            return array(
                'result'    => 'success',
                'redirect'    =>  $redirect_url
            );
        } elseif ($sagepay_response['Status'] == "3DAUTH") {
            if ($sagepay_response['3DSecureStatus'] == 'OK') {
                if (isset($sagepay_response['ACSURL']) && isset($sagepay_response['MD'])) {
                    WC()->session->set('sagepay_acsurl', $sagepay_response['ACSURL']);
                    WC()->session->set('sagepay_pareq', $sagepay_response['PAReq']);
                    WC()->session->set('sagepay_md', $sagepay_response['MD']);
                    WC()->session->set( "Complete3d", $order->get_checkout_payment_url( true ) );

                    $redirect = $order->get_checkout_payment_url(true);

                    return array(
                        'result'    => 'success',
                        'redirect'    => $redirect
                    );
                }
            }
        } else {
            if (isset($sagepay_response['StatusDetail'])) {
                wc_add_notice(sprintf(__('Transaction Failed. %s - %s', 'ag-sagepay-direct'), $sagepay_response['Status'], $sagepay_response['StatusDetail']), 'error');
            } else {
                wc_add_notice(sprintf(__('Transaction Failed with %s - unknown error.', 'ag-sagepay-direct'), $sagepay_response['Status']), 'error');
            }
        }
    } else {
        wc_add_notice(__('Gateway Error. Please Notify the Store Owner about this error.', 'ag-sagepay-direct'), 'error');
    }
}

/*-----------------------------------------------------------------------------------*/
/*  Setting receipt_page
/*-----------------------------------------------------------------------------------*/
public function receipt_page($order) {
    global $woocommerce;
    echo '<p>'.__('Thank you for your order, Please click button below to Authenticate your card.', 'ag-sagepay-direct').'</p>';
    echo $this->generate_sagepaydirect_form($order);
}

public function process_refund( $order_id, $amount = NULL, $reason = '' ) {

    	$order 			   = new WC_Order( $order_id );

			$VendorTxCode  = 'Refund-' . $order_id . '-' . time();

			// API Request for refund
			$api_request 	 = 'VPSProtocol=' . '3.00';
			$api_request 	.= '&TxType=REFUND';
			$api_request  .= '&Vendor=' . $this->vendor;
			$api_request 	.= '&VendorTxCode=' . $VendorTxCode;
			$api_request  .= '&Amount=' . urlencode( $amount );
			$api_request 	.= '&Currency=' . get_woocommerce_currency();
			$api_request 	.= '&Description=Refund for order ' . $order_id;
			$api_request	.= '&RelatedVPSTxId=' . get_post_meta( $order_id, 'VPSTxId', true );
			$api_request	.= '&RelatedVendorTxCode=' . get_post_meta( $order_id, 'VendorTxCode', true );
			$api_request	.= '&RelatedSecurityKey=' . get_post_meta( $order_id, 'SecurityKey', true );
			$api_request	.= '&RelatedTxAuthNo=' . get_post_meta( $order_id, 'TxAuthNo', true );

      if( is_ssl()) {
        $environment_url = ( "FALSE" == $environment )
                   ? 'https://live.sagepay.com/gateway/service/refund.vsp'
                   : 'https://test.sagepay.com/gateway/service/refund.vsp';
      } else {
        $environment_url = 'https://test.sagepay.com/gateway/service/refund.vsp';
      }

			$result = wp_remote_post($environment_url, array(
        'method' => 'POST',
        'timeout'     => 6,
        'redirection' => 5,
        'body' => $api_request,
      ));

      $sagepay_response = array();
      $lines = preg_split('/\r\n|\r|\n/', $result['body']);
      foreach ($lines as $line) {
        $key_value = preg_split('/=/', $line, 2);
        if (count($key_value) > 1) {
          $sagepay_response[trim($key_value[0])] = trim($key_value[1]);
        }
      }


			if ( $sagepay_response['Status'] == "OK" ) {
        $order->add_order_note( __('Refund successful', 'ag-sagepay-direct') . '<br />' .
          __('Refund Amount : ', 'ag-sagepay-direct') . $amount . '<br />' .
          __('Refund Reason : ', 'ag-sagepay-direct') . $reason . '<br />'
      );
      return true;

			} else {

				$order->add_order_note( __('Refund failed', 'ag-sagepay-direct') . '<br />'  );

			}

}


/**
 * WOOP WOOP Successful Payment!! ;)
 **/
public function successful_request() {
      global $woocommerce;

      if (isset($_REQUEST['MD']) && isset($_REQUEST['PaRes'])) {
          $order = new WC_Order(WC()->session->get('sagepay_oid'));
          $request_array = array(
            'MD' => $_REQUEST['MD'],
            'PARes' => $_REQUEST['PaRes'],
            'VendorTxCode' => WC()->session->get('sagepay_vtc'),
          );

          $request = http_build_query($request_array);

          // Are we testing right now or is it a real transaction
          $environment = ( $this->environment == "yes" ) ? 'TRUE' : 'FALSE';

          // Decide which URL to post to
          if( is_ssl()) {
          	$environment_url = ( "FALSE" == $environment )
          						 ? 'https://live.sagepay.com/gateway/service/direct3dcallback.vsp'
          						 : 'https://test.sagepay.com/gateway/service/direct3dcallback.vsp';
          } else {
            $environment_url = 'https://test.sagepay.com/gateway/service/direct3dcallback.vsp';
          }

          $result = wp_remote_post($environment_url, array(
            'body' => $request,
            'method' => 'POST',
            'sslverify' => false
          ));

          if (!is_wp_error($result) && $result['response']['code'] >= 200 && $result['response']['code'] < 300) {
              $sagepay_response = array();
              $lines = preg_split('/\r\n|\r|\n/', $result['body']);
              foreach ($lines as $line) {
                  $key_value = preg_split('/=/', $line, 2);
                  if (count($key_value) > 1) {
                      $sagepay_response[trim($key_value[0])] = trim($key_value[1]);
                  }
              }

              if (isset($sagepay_response['Status'])) {
                  update_post_meta($order->id, 'Status', $sagepay_response['Status']);
              }
              if (isset($sagepay_response['StatusDetail'])) {
                  update_post_meta($order->id, 'StatusDetail', $sagepay_response['StatusDetail']);
              }
              if (isset($sagepay_response['VPSTxId'])) {
                  update_post_meta($order->id, 'VPSTxId', $sagepay_response['VPSTxId']);
              }
              if (isset($sagepay_response['CAVV'])) {
                  update_post_meta($order->id, 'CAVV', $sagepay_response['CAVV']);
              }
              if (isset($sagepay_response['SecurityKey'])) {
                  update_post_meta($order->id, 'SecurityKey', $sagepay_response['SecurityKey']);
              }
              if (isset($sagepay_response['TxAuthNo'])) {
                  update_post_meta($order->id, 'TxAuthNo', $sagepay_response['TxAuthNo']);
              }
              if (isset($sagepay_response['AVSCV2'])) {
                  update_post_meta($order->id, 'AVSCV2', $sagepay_response['AVSCV2']);
              }
              if (isset($sagepay_response['AddressResult'])) {
                  update_post_meta($order->id, 'AddressResult', $sagepay_response['AddressResult']);
              }
              if (isset($sagepay_response['PostCodeResult'])) {
                  update_post_meta($order->id, 'PostCodeResult', $sagepay_response['PostCodeResult']);
              }
              if (isset($sagepay_response['CV2Result'])) {
                  update_post_meta($order->id, 'CV2Result', $sagepay_response['CV2Result']);
              }
              if (isset($sagepay_response['3DSecureStatus'])) {
                  update_post_meta($order->id, '3DSecureStatus', $sagepay_response['3DSecureStatus']);
              }
              if (WC()->session->get('sagepay_vtc') != '') {
                  update_post_meta($order->id, 'VendorTxCode', WC()->session->get('sagepay_vtc'));
              }

              if ($sagepay_response['Status'] == "OK" || $sagepay_response['Status'] == "REGISTERED" || $sagepay_response['Status'] == "AUTHENTICATED") {
                  $order->add_order_note(__('SagePay Direct Payment Completed.', 'ag-sagepay-direct'));
                  $order->payment_complete();
                  $redirect_url = $this->get_return_url($order);
                  wp_redirect($redirect_url);
                  exit();
              } elseif ($sagepay_response['Status'] == "3DAUTH") {
                  if ($sagepay_response['3DSecureStatus'] == 'OK') {
                      if (isset($sagepay_response['ACSURL']) && isset($sagepay_response['MD'])) {
                          WC()->session->set('sagepay_acsurl', $sagepay_response['ACSURL']);
                          WC()->session->set('sagepay_pareq', $sagepay_response['PAReq']);
                          WC()->session->set('sagepay_md', $sagepay_response['MD']);
                          $redirect = $order->get_checkout_payment_url(true);
                          wp_redirect($redirect);
                          exit();
                      }
                  }
              } else {
                  if (isset($sagepay_response['StatusDetail'])) {
                      wc_add_notice(sprintf(__('Transaction Failed. %s - %s', 'ag-sagepay-direct'), $sagepay_response['Status'], $sagepay_response['StatusDetail']), 'error');
                      $get_checkout_url = apply_filters('woocommerce_get_checkout_url', WC()->cart->get_checkout_url());
                      wp_redirect($get_checkout_url);
                      exit();
                  } else {
                      wc_add_notice(sprintf(__('Transaction Failed with %s - unknown error.', 'ag-sagepay-direct'), $sagepay_response['Status']), 'error');
                      $get_checkout_url = apply_filters('woocommerce_get_checkout_url', WC()->cart->get_checkout_url());
                      wp_redirect($get_checkout_url);
                      exit();
                  }
              }
          } else {
              wc_add_notice(__('Gateway Error. Please Notify the Store Owner about this error.', 'ag-sagepay-direct'), 'error');
              $get_checkout_url = apply_filters('woocommerce_get_checkout_url', WC()->cart->get_checkout_url());
              wp_redirect($get_checkout_url);
              exit();
          }
      }
  }
}
